/*
 * File name: pctspn_som.h
 * Date:      2013/10/13 09:02
 * Author:    Jan Faigl
 */

#ifndef __PCTSPN_SOM_H__
#define __PCTSPN_SOM_H__


#include <vector>

#include <imr_config.h>
#include <imr_algorithm.h>
#include <text_format_loader.h>

#include "coords.h"
#include "station.h"

namespace som {

   class CPCTSPNSom : public imr::CAlgorithm {
      typedef imr::CAlgorithm Base;
      typedef std::vector<int> IntVector;

      public:
      typedef imr::loader::CTextFormatLoader Loader;

      static imr::CConfig& getConfig(imr::CConfig& config);

      static std::string getName(void) { return "pctspsom"; }

      CPCTSPNSom(imr::CConfig& config, const std::string& problemFile, bool loadPenalty = false);
      ~CPCTSPNSom();

      std::string getRevision(void);

      std::string getMethod(void) { return getName(); }

      void setTimers(const imr::CTimerN& load, const imr::CTimerN& init);

      std::string getVersion(void);

      void solve(void);

      public:
      struct SNeuron {
	 SNeuron* prev;
	 SNeuron* next;
	 Coords coords;
	 Coords alternateGoal;
	 int stationOnTourStep;
	 int stationOnTour; 

	 SNeuron() {
	    clear();
	 }

	 SNeuron(const Coords& pt) : coords(pt) {
	    clear();
	 } 

	 ~SNeuron() {
	 }

	 void setPoint(const Coords& pt) {
	    coords = pt;
	    clear();
	 }

	 void clear(void) { 
	    stationOnTour = -1;
	    stationOnTourStep = -1;
	 }

	 void adapt(const Coords& pt, const double beta) {
	    coords.x += beta * (pt.x - coords.x);
	    coords.y += beta * (pt.y - coords.y);
	 }

	 inline bool isInhibited(int step) { return stationOnTourStep >= step; }
      };
      protected:

      //# methods from CAlgorithm
      void load(void);
      void initialize(void);
      void after_init(void);
      void iterate(int iter);
      void save(void);
      void release(void);
      void visualize(void);
      void defineResultLog(void);
      void fillResultRecord(int iter, double length, double penalty, int steps);

      void allocate_neurons(void); // allocate neurons
      void deallocate_neurons(void); // release neurons
      void initialize_neurons(void); // place neurons around the first station

      /// ----------------------------------------------------------------------------
      /// @return the added neuron
      /// ----------------------------------------------------------------------------
      SNeuron* insertNeuron(SNeuron* cur, SNeuron* neuron);
      void removeNeuron(SNeuron* neuron);

      double refine(int step, double errorMax);

      private:
      SNeuron* selectWinner(int step, SStation* station, double& errorToCity);
      void adapt(SNeuron* neuron, const int step);

      void drawRing(int step);

      void drawPath(void);

      void savePic(int step, bool detail = false, const std::string& dir_suffix = "");
      void releaseStations(void);

      void ring_regenerate(int step);

      void load_stations_coords(const std::string& filename);
      void load_stations_penalty(const std::string& filename);


      CoordsVector& get_ring_path(int step, CoordsVector& path) const;
      double get_path_length(const CoordsVector& path, bool closed = true) const;
      std::vector<bool>& get_path_covered_stations(const CoordsVector& path, std::vector<bool>& covered, bool closed = true) const;
      double get_path_penalty(const CoordsVector& path, int* coveredStations = 0) const;

      void processPath(void);

      /// ----------------------------------------------------------------------------
      /// Variables
      /// ----------------------------------------------------------------------------
      protected:
      IntVector permutation;

      double mi;

      double G; //variable
      int n; //number stations
      int d; //number neighborhood neurons

      SNeuron* neurons;
      int nbrNeurons;

      imr::CTimerN loadTimer;
      imr::CTimerN initTimer;
      bool savePicEnabled;
      std::vector<bool> coveredStations;
      const double GAIN_DECREASING_RATE;
      const double NEIGHBORHOOD_FACTOR;

      const double COMM_RADIUS;
      const double COMM_RADIUS2;
      const double COMM_RADIUS_SMALLER;
      const bool SAVE_RESULTS;
      const bool SAVE_INFO;
      const bool SAVE_SETTINGS;
      const bool PROCESS_PATH;

      StationPtrVector stations;
      bool drawNeurons;
      bool drawAlternateGoals;
      bool drawWinners;
      bool drawCommRadius;
      bool drawWinnerCommRadius;
      bool drawTourRepresentedByRing;

      imr::gui::CShape shapeRing;
      imr::gui::CShape shapeNeuron;
      imr::gui::CShape shapeWinner;
      imr::gui::CShape shapeCommRadius;
      imr::gui::CShape shapeCommRadiusCoverage;
      imr::gui::CShape shapeAlternateGoal;
      imr::gui::CShape shapeTourRepresentedByRing;
      bool drawRingIter;
      bool drawRingEnabled;
      CoordsVector finalPath;

      bool drawCoverageEnable;
      bool drawCoveredEnable;
      bool drawCoverageLastEnable;
      imr::gui::CShape shapeCoverage;
   };

} //end namespace som

#endif

/* end of pctspn_som.h */
